
#include <string.h>
#include "globdef.h"
#include "conf.h"
#include "uidef.h"
#include "padef.h"
#include "thrdef.h"
#include "sigdef.h"

#if PORTAUDIO_PRESENT == 1
PaStream *out_stream ;
#if(OSNUM == OS_FLAG_LINUX)
#if JACK_PRESENT == 1
const char linrad_name[8] ="linrad";
PaError PaJack_SetClientName(const char *name);
#endif
#endif
#endif

#if PORTAUDIO_PRESENT == 1
int daout_workload_reset;

/// This routine will be called by the PortAudio engine when audio is needed.
/// It may be called at interrupt level on some machines so don't do anything
/// that could mess up the system like calling malloc() or free().

extern int playCallback( const void *inputBuffer, void *outputBuffer,
                         unsigned long framesToProcess,
                         const PaStreamCallbackTimeInfo* timeInfo,
                         PaStreamCallbackFlags statusFlags,
                         void *userData )
{
// Prevent unused variable warnings.
(void) inputBuffer;
(void) timeInfo;
(void) statusFlags;
(void) userData;
int samps;
if(portaudio_active_flag == FALSE)
  {
  lirerr(701400);
  return paAbort;
  }
if(snd[RXDA].block_frames != (int)framesToProcess)
  {
  lirerr(520673);
  return paAbort;
  }
if(thread_command_flag[THREAD_RX_OUTPUT]==THRFLAG_IDLE ||
           thread_status_flag[THREAD_RX_OUTPUT]==THRFLAG_IDLE)goto idle;
samps=(daout_pa-daout_px+daout_size)&daout_bufmask;
if(daout_workload_reset!=workload_reset_flag)
  {
  min_daout_samps=baseband_size;
  daout_workload_reset=workload_reset_flag;
  }
if(audio_dump_flag == 1)goto idle;
if( samps > snd[RXDA].block_bytes)
  {
  memcpy(outputBuffer,&daout[daout_px],snd[RXDA].block_bytes);
  daout_px=(daout_px+snd[RXDA].block_bytes)&daout_bufmask;
  samps-=snd[RXDA].block_bytes;
  if(min_daout_samps > samps)min_daout_samps=samps;
  }
else
  {
  min_daout_samps=0;
idle:;  
  memset(outputBuffer, 0, snd[RXDA].block_bytes);
  }
return paContinue;
}
#endif

void portaudio_stop(void)
{
#if PORTAUDIO_PRESENT == 1
if(portaudio_active_flag == FALSE)
  {
  lirerr(701401);
  return;
  }
Pa_Terminate();
#endif
}

void open_portaudio_rxda(void)
{
#if PORTAUDIO_PRESENT == 1
PaStreamParameters  rxdaParameters;
const PaStreamInfo *streamInfo;
PaError pa_err;
if(portaudio_active_flag == FALSE)
  {
  lirerr(701401);
  return;
  }
// Set sample format
if(rx_daout_bytes == 1)
  {
  rxdaParameters.sampleFormat = paUInt8; // 1 byte samples
  }
else
  {  
  rxdaParameters.sampleFormat = paInt16; // 2 byte samples
  }
rxdaParameters.device       = ui.rx_dadev_no;
rxdaParameters.channelCount = rx_daout_channels;
rxdaParameters.suggestedLatency = 1/snd[RXDA].interrupt_rate ;



rxdaParameters.hostApiSpecificStreamInfo = NULL;
pa_err = Pa_IsFormatSupported( NULL,&rxdaParameters, genparm[DA_OUTPUT_SPEED] );
if( pa_err != paFormatIsSupported )
  {
  ui.rx_dadev_no = -2;
  goto start_pa_end;
  }
// Open and Start audio I/O stream.
pa_err = Pa_OpenStream( 
              &out_stream,
              NULL,                               //no input
              &rxdaParameters,
              (double)(genparm[DA_OUTPUT_SPEED]), //SAMPLE_RATE,
              snd[RXDA].block_frames,             //frames per buffer
              paClipOff,                          //no clipping
              playCallback,
              NULL );                             //no userdata
if( pa_err != paNoError )
  {
  ui.rx_dadev_no = -2;
  goto start_pa_end;
  }
pa_err = Pa_StartStream( out_stream );
if( pa_err != paNoError )
  {
  ui.rx_dadev_no = -2;
  goto start_pa_end;
  }
streamInfo = Pa_GetStreamInfo(out_stream);
DEB"\nLATENCY %f\n",streamInfo->outputLatency);

start_pa_end:;
// set an arbitrary positive value.
rx_audio_out=1;
snd[RXDA].tot_bytes=snd[RXDA].no_of_blocks*snd[RXDA].block_bytes;
daout_workload_reset=workload_reset_flag;
#else
lirerr(700124);
#endif
}

void close_portaudio_rxda(void)
{
#if PORTAUDIO_PRESENT == 1
PaError pa_err;
pa_err = Pa_CloseStream( out_stream );
if( pa_err != paNoError )
  {
  lirerr(1168);
  }
#else
lirerr(700123);
#endif
}

void portaudio_startstop(void)
{
#if PORTAUDIO_PRESENT == 1
PaErrorCode err;
if( (ui.use_alsa&PORTAUDIO_RXIN) != 0 ||
    (ui.use_alsa&PORTAUDIO_RX_OUT) != 0 ||
    (ui.use_alsa&PORTAUDIO_TXIN) != 0 ||
    (ui.use_alsa&PORTAUDIO_TX_OUT) != 0 )
  {
  if(portaudio_active_flag == FALSE)
    {
#if PORTAUDIO_PRESENT == 1
#if(OSNUM == OS_FLAG_LINUX)
#if JACK_PRESENT == 1
    PaJack_SetClientName(linrad_name);
#endif
#endif
    err=Pa_Initialize();
    if(err < 0)
      {
      lirerr(847297);
      return;
      }      
DEB"\nPortAudio ver %d \n%s\n", Pa_GetVersion(), Pa_GetVersionText());
#endif
    }
  portaudio_active_flag=TRUE;
  }
else
  {  
  if(portaudio_active_flag == TRUE)
    {
    err=Pa_Terminate();
    if(err < 0)
      {
      lirerr(847298);
      return;
      }
    }
  portaudio_active_flag=FALSE;
  }
#else
portaudio_active_flag=FALSE;
#endif
}

void set_portaudio(int sound_type, int *line)
{
(void) sound_type;
(void) line;
#if PORTAUDIO_PRESENT == 1
int i;
PaError pa_err;
char s[128];
char devflag[MAX_PA_DEVICES];
double devmaxStandardSampleRate[MAX_PA_DEVICES];
double devminStandardSampleRate[MAX_PA_DEVICES];
int devmaxBytes[MAX_PA_DEVICES];
int devminBytes[MAX_PA_DEVICES];
int devmaxInputChannels[MAX_PA_DEVICES];
int devminInputChannels[MAX_PA_DEVICES];
int devmaxOutputChannels[MAX_PA_DEVICES];
int devminOutputChannels[MAX_PA_DEVICES];
int j, numDevices,valid_dev_cnt;
const PaDeviceInfo *deviceInfo;
PaStreamParameters inputParameters;
PaStreamParameters outputParameters;
int id, minBytes, maxBytes;
int maxInputChannels, maxOutputChannels;
int minInputChannels, minOutputChannels;
double maxStandardSampleRate;
double minStandardSampleRate;
// negative terminated  list
static double standardSampleRates[] = {8000.0, 9600.0, 
        11025.0, 12000.0, 16000.0, 22050.0, 24000.0, 32000.0,
        44100.0, 48000.0, 88200.0, 96000.0, 192000.0, -1};
// *******************************************************************
SNDLOG"\nTESTING PORTAUDIO DEVICES FOR %s\n",sndtype[sound_type]);
SNDLOG"\nPortAudio ver %d \n%s\n", Pa_GetVersion(),Pa_GetVersionText());
valid_dev_cnt=0;
minBytes=0;
maxBytes=0;
minOutputChannels=0;
minInputChannels=0;
maxOutputChannels=0;
minStandardSampleRate=0;
maxStandardSampleRate=0;
clear_screen();
line[0]=4;
numDevices=Pa_GetDeviceCount();
if(numDevices < 0)
  {
  lirerr(1166);
  return;
  }
if(numDevices == 0)
  {
  lir_text(5,2,"No portaudio device detected for testing.");
  SNDLOG"\nNo portaudio device detected for testing.");
  lir_text(5,4,press_any_key);
  await_keyboard();
  goto end_portaudiotest;
  }
  else
   {
   SNDLOG"\nNumber of portaudio devices to be tested=%i\n",numDevices);
   }
fflush(sndlog);
settextcolor(14);
sprintf(s,"List of valid portaudio devices for %s",sndtype[sound_type]);
lir_text(20,0,s);
if(numDevices > MAX_PA_DEVICES)
  {
  lir_text(0,line[0],"TOO MANY DEVICES. List truncated.");
  numDevices=MAX_PA_DEVICES;
  }
settextcolor(7);
// headings
lir_text(85,line[0],"SAMPLE RATE");
lir_text(101,line[0],"CHANNELS");
lir_text(116,line[0],"BYTES");
line[0]++;
lir_text(12,line[0],"ID         NAME");
lir_text(64,line[0],"HOST-API");
lir_text(87,line[0],"MIN/MAX");
lir_text(101,line[0],"MIN/MAX");
lir_text(115,line[0],"MIN/MAX");
line[0]++;
lir_text(12,line[0],"--  ------------------------");
lir_text(64,line[0],"-------");
lir_text(84,line[0],"-------------");
lir_text(101,line[0],"-------");
lir_text(115,line[0],"-------");
line[0]++;
lir_refresh_screen();

// list valid devices
for(id=0; id < (numDevices); id++)
  {
  devflag[id]=0;
  deviceInfo = Pa_GetDeviceInfo( id );
  maxInputChannels= deviceInfo->maxInputChannels;
  maxOutputChannels= deviceInfo->maxOutputChannels;
#if (OSNUM == OS_FLAG_WINDOWS)
// bypass bug in Juli@ ASIO driver: */
// this driver hangs after a Pa_IsFormatSupported call */
  i = strncmp(deviceInfo->name, "ASIO 2.0 - ESI Juli@", 19);
  if (i == 0)
    {
    minStandardSampleRate=44100;
    maxStandardSampleRate=192000;
    minBytes=1;
    maxBytes=4;
    maxInputChannels= deviceInfo->maxInputChannels;
    maxOutputChannels= deviceInfo->maxOutputChannels;
    goto endpoll;
    }
#endif
  SNDLOG"\n\nid=%d device=%s hostapi=%s max_inp_chan=%d max_outp_chan=%d", id, deviceInfo->name,
          Pa_GetHostApiInfo( deviceInfo->hostApi )->name, maxInputChannels, maxOutputChannels);
  fflush(sndlog);
  if (maxOutputChannels==0 && 
         (sound_type==RXDA || sound_type==TXDA))goto next_id;

  if (maxInputChannels==0 && 
         (sound_type==RXAD || sound_type==TXAD))goto next_id; 
  maxStandardSampleRate=0;
  minStandardSampleRate=0;
// Investigate device capabilities.
// Start by investigating the RDWR capabilities with 16 bit data.
  inputParameters.device = id;
  inputParameters.channelCount = deviceInfo->maxInputChannels;
  inputParameters.sampleFormat = paInt16;
  inputParameters.suggestedLatency = 0;
  inputParameters.hostApiSpecificStreamInfo = NULL;
  outputParameters.device = id;
  outputParameters.channelCount = deviceInfo->maxOutputChannels;
  outputParameters.sampleFormat = paInt16;
  outputParameters.suggestedLatency = 0;
  outputParameters.hostApiSpecificStreamInfo = NULL;
  i=0;
  while(standardSampleRates[i] > 0 && minStandardSampleRate==0)
    {
    SNDLOG"\nTesting %d %f",i,standardSampleRates[i]);
    if(sound_type==RXDA || sound_type==TXDA)
      {
      pa_err=Pa_IsFormatSupported(
                          NULL, &outputParameters, standardSampleRates[i] );
      }
    else
      {  
      pa_err=Pa_IsFormatSupported(
                  &inputParameters, NULL, standardSampleRates[i] );
      }
    SNDLOG"  err=%d  %s",pa_err, Pa_GetErrorText(pa_err));
    fflush(sndlog);  
    if(pa_err == paDeviceUnavailable)goto next_id;
    if(pa_err == paInvalidDevice)goto next_id;
    if(pa_err == paFormatIsSupported )
      { 
      minStandardSampleRate=standardSampleRates[i];
      }
    i++;
    }
  if(minStandardSampleRate == 0)goto next_id;
  j=i;
  while(standardSampleRates[i] > 0 )i++;
  i--;
  while(i >= j && maxStandardSampleRate==0)
    {
    SNDLOG"\nTesting %d %f",i,standardSampleRates[i]);
    if(sound_type==RXDA || sound_type==TXDA)
      {
      pa_err=Pa_IsFormatSupported(
                          NULL, &outputParameters, standardSampleRates[i] );
      }
    else
      {  
      pa_err=Pa_IsFormatSupported(
                  &inputParameters, NULL, standardSampleRates[i] );
      }
    SNDLOG"  err=%d  %s",pa_err, Pa_GetErrorText(pa_err));
    fflush(sndlog);   
    if(pa_err == paDeviceUnavailable)goto next_id;
    if(pa_err == paInvalidDevice)goto next_id;
    if( pa_err == paFormatIsSupported )
      { 
      maxStandardSampleRate=standardSampleRates[i];
      }
    i--;
    }
// check maxStandardSampleRate 
  if(maxStandardSampleRate==0 && 
         (minStandardSampleRate != 0))
    {
    maxStandardSampleRate= minStandardSampleRate;
    }
// check min and max bytes
  minBytes=2;
  maxBytes=2;
  if(sound_type==RXDA || sound_type==TXDA)
    {
    outputParameters.sampleFormat = paUInt8;
    pa_err=Pa_IsFormatSupported(
                          NULL, &outputParameters, maxStandardSampleRate );
    }
  else
    {  
    inputParameters.sampleFormat = paUInt8;
    pa_err=Pa_IsFormatSupported(
                  &inputParameters, NULL, maxStandardSampleRate );
    }
  if( pa_err == paFormatIsSupported )
    { 
    minBytes=1;
    }
  SNDLOG"\nminBytes=%i", minBytes); 
  fflush(sndlog);

  if(sound_type==RXDA || sound_type==TXDA)
    {
    outputParameters.sampleFormat = paInt32;
    pa_err=Pa_IsFormatSupported(
                          NULL, &outputParameters, maxStandardSampleRate );
    }
  else
    {  
    inputParameters.sampleFormat = paInt32;
    pa_err=Pa_IsFormatSupported(
                  &inputParameters, NULL, maxStandardSampleRate );
    }
  if( pa_err == paFormatIsSupported )
    { 
    maxBytes=4;
    }
  SNDLOG"\nmaxBytes=%i", maxBytes); 
  fflush(sndlog);
#if (OSNUM == OS_FLAG_WINDOWS)
endpoll:;
#endif
//check min channelCount
  if(sound_type==RXDA || sound_type==TXDA)
    {
      outputParameters.channelCount = 1;
      outputParameters.sampleFormat = paInt16;
      pa_err=paFormatIsSupported+32000;
      while(pa_err != paFormatIsSupported &&
           ( outputParameters.channelCount < (maxOutputChannels+1)) )
        {
        pa_err=Pa_IsFormatSupported(
                          NULL, &outputParameters, maxStandardSampleRate );
        outputParameters.channelCount++;
        }
      if( pa_err == paFormatIsSupported )
        {
        minOutputChannels=outputParameters.channelCount-1;
        SNDLOG"\nminOutputChannels=%i", minOutputChannels); 
        fflush(sndlog); 
        }
      else
        {
        SNDLOG"\nminOutputChannels=%i samplerate=%6.0f err=%d  %s"
           , (outputParameters.channelCount-1), maxStandardSampleRate, pa_err, Pa_GetErrorText(pa_err));
        fflush(sndlog);
        goto next_id;
        }
    }
  else
    { 
      inputParameters.channelCount = 1;
      inputParameters.sampleFormat = paInt16;
      pa_err=paFormatIsSupported+32000;
      while(pa_err != paFormatIsSupported &&
           ( inputParameters.channelCount < (maxInputChannels+1)) )
        {
        pa_err=Pa_IsFormatSupported(
                  &inputParameters, NULL, maxStandardSampleRate );
        inputParameters.channelCount++;
        }
      if( pa_err == paFormatIsSupported )
        {
        minInputChannels=inputParameters.channelCount-1;
        SNDLOG"\nminInputChannels=%i", minInputChannels); 
        fflush(sndlog);
        }
      else
        {
        SNDLOG"\nminInputChannels=%i samplerate=%6.0f err=%d  %s"
            ,( inputParameters.channelCount-1), maxStandardSampleRate, pa_err, Pa_GetErrorText(pa_err));
        fflush(sndlog);
        goto next_id;
        }
     }
//display result to screen
  if(sound_type==RXDA || sound_type==TXDA)
    {
    sprintf(s,"%2d/%2d",minOutputChannels,maxOutputChannels);
    }
  else
    {
    sprintf(s,"%2d/%2d",minInputChannels, maxInputChannels);
    }
  lir_text(101,line[0],s);
  sprintf(s,"%02d  %s", id, deviceInfo->name);
  lir_text(12,line[0],s);
  sprintf(s," %s", Pa_GetHostApiInfo( deviceInfo->hostApi )->name );
  lir_text(63,line[0],s);
  sprintf(s,"%6.0f/%6.0f",minStandardSampleRate, maxStandardSampleRate);
  lir_text(83,line[0],s);
  sprintf(s,"%i/%i",minBytes, maxBytes);
  lir_text(117,line[0],s);
  line[0]++;
  lir_refresh_screen();
//save values
  devmaxStandardSampleRate[id]=maxStandardSampleRate;
  devminStandardSampleRate[id]=minStandardSampleRate;
  devmaxBytes[id]=maxBytes;
  devminBytes[id]=minBytes;
  devmaxInputChannels[id]=maxInputChannels;
  devminInputChannels[id]=minInputChannels;
  devmaxOutputChannels[id]=maxOutputChannels;
  devminOutputChannels[id]=minOutputChannels;
  devflag[id]=1;
  
  valid_dev_cnt++;
  next_id:;
  }
SNDLOG"\n\nEND TESTING PORTAUDIO DEVICES\n\n");
if (valid_dev_cnt !=0)
  {
  line[0]++;
  settextcolor(14);
  sprintf(s,"Select device for %s from list",sndtype[sound_type]);
  lir_text(10,line[0],s);
  SNDLOG"\n%s",s);
  settextcolor(7);
  gtdev:;
  id=lir_get_integer(strlen(s)+12,line[0], 2, 0, numDevices-1);
  SNDLOG"\nUser selected %i\n",id);
  if(kill_all_flag)return;
  if(devflag[id] == 0)goto gtdev;
  line[0]++;
  switch (sound_type)
    {
    case RXDA:
    ui.rx_dadev_no=id;
    ui.rx_max_da_speed=devmaxStandardSampleRate[id];
    ui.rx_min_da_speed=devminStandardSampleRate[id];
    ui.rx_min_da_bytes=devminBytes[id];
    ui.rx_max_da_bytes=devmaxBytes[id];
    ui.rx_min_da_channels= devminOutputChannels[id];
    ui.rx_max_da_channels=devmaxOutputChannels[id];
// Change the limits for RXDA if they are too generous or seem incorrect.
    if(ui.rx_max_da_bytes > 2)ui.rx_max_da_bytes=2;
    if(ui.rx_min_da_channels >=  2)ui.rx_min_da_channels=2;
           else ui.rx_min_da_channels=1;
    if(ui.rx_max_da_channels <  2 )ui.rx_max_da_channels=1;
           else ui.rx_max_da_channels=2;
    break;
    } 
  }
else
  {
  line[0]++;
  settextcolor(14);
  lir_text(10,line[0],"No valid portaudio device detected ");
  SNDLOG"\nNo valid portaudio device detected ");
  line[0]++;
  lir_text(10,line[0],press_any_key);
  settextcolor(7);
  await_keyboard();
  } 
end_portaudiotest:;
#endif
}

